@@ -516,14 +516,14 @@ static gint open_command(assuan_context_t ctx, gchar *line)
 
     if (!filename || !*filename) {
 	pth_cleanup_pop(1);
-	return send_error(ctx, EPWMD_COMMAND_SYNTAX);
+	return send_error(ctx, GPG_ERR_SYNTAX);
     }
 
     log_write2("ARGS=\"%s\" %s", filename, req[1] ? "<passphrase>" : "");
 
     if (valid_filename(filename) == FALSE) {
 	pth_cleanup_pop(1);
-	return send_error(ctx, EPWMD_INVALID_FILENAME);
+	return send_error(ctx, GPG_ERR_INV_VALUE);
     }
 
     if (client->state == STATE_OPEN)
@@ -1333,7 +1333,8 @@ static gint save_command(assuan_context_t ctx, gchar *line)
 
     memset(client->crypto->key, '!', hashlen);
 
-    if (get_key_file_double(client->filename, "iterations") <= 0L)
+    if (get_key_file_double(client->filename, "iterations") <= 0L &&
+	    (!line || !*line))
 	goto done;
 
     if (!line || !*line) {
@@ -1386,9 +1387,26 @@ static gint save_command(assuan_context_t ctx, gchar *line)
 	    cached = TRUE;
 	}
     }
-    else
+    else {
+	if (get_key_file_double(client->filename, "iterations") <= 0L) {
+	    guint64 iter = (guint64)get_key_file_double(NULL, "iterations");
+
+	    if (iter <= 0L)
+		iter = 1;
+
+	    MUTEX_LOCK(&rcfile_mutex);
+	    g_key_file_set_double(keyfileh, client->filename, "iterations", iter);
+	    MUTEX_UNLOCK(&rcfile_mutex);
+	    client->opts |= OPT_ITERATIONS;
+	    gpg_error_t rc = send_status(ctx, STATUS_CONFIG, NULL);
+
+	    if (rc)
+		return send_error(ctx, rc);
+	}
+
 	gcry_md_hash_buffer(GCRY_MD_SHA256, client->crypto->key, line,
 		strlen(line));
+    }
 
 done:
     return save_command_finalize(ctx, client->crypto->key, cached);
@@ -1409,7 +1427,7 @@ static gint delete_command(assuan_context_t ctx, gchar *line)
 	req = split_input_line(line, " ", -1);
 
     if (!req || !*req)
-	return send_error(ctx, EPWMD_COMMAND_SYNTAX);
+	return send_error(ctx, GPG_ERR_SYNTAX);
 
     n = find_root_element(client->doc, &req, &rc, NULL, 0, FALSE);
 
@@ -1469,12 +1487,12 @@ static gint store_command_finalize(gpointer data, gint assuan_rc, guchar *line,
     xfree(line);
 
     if (!req || !*req)
-	return EPWMD_COMMAND_SYNTAX;
+	return GPG_ERR_SYNTAX;
 
 again:
     n = find_root_element(client->doc, &req, &rc, NULL, 0, FALSE);
 
-    if (rc && rc == EPWMD_ELEMENT_NOT_FOUND) {
+    if (rc && rc == GPG_ERR_ELEMENT_NOT_FOUND) {
 	rc = new_root_element(client->doc, *req);
 
 	if (rc) {
@@ -1563,11 +1581,22 @@ gpg_error_t do_assuan_command(assuan_context_t ctx,
     pth_cleanup_push(cleanup_cancel_cb, tid);
     to = to <= 0 ? DEFAULT_KEEPALIVE_TO : to;
     ev = pth_event(PTH_EVENT_TID|PTH_UNTIL_TID_DEAD, tid);
-    tev = pth_event(PTH_EVENT_TIME, pth_timeout(to, 0));
+    tev = to ? pth_event(PTH_EVENT_TIME, pth_timeout(to, 0)) : NULL;
     ev = pth_event_concat(ev, tev, NULL);
     pth_cleanup_push(cleanup_ev_cb, ev);
     pth_yield(tid);
     pth_wait(ev);
+
+    if (tev) {
+	st = pth_event_status(tev);
+
+	if (st == PTH_STATUS_OCCURRED) {
+	    pth_cleanup_pop(1);
+	    pth_cleanup_pop(1);
+	    return GPG_ERR_TIMEOUT;
+	}
+    }
+
     st = pth_event_status(ev);
 
     if (st == PTH_STATUS_FAILED) {
@@ -1581,16 +1610,6 @@ gpg_error_t do_assuan_command(assuan_context_t ctx,
 	rc = *(gpg_error_t *)p;
 	g_free(p);
     }
-    else {
-	st = pth_event_status(tev);
-
-	if (st == PTH_STATUS_OCCURRED) {
-	    pth_cancel(tid);
-	    pth_join(tid, &p);
-	    g_free(p);
-	    rc = GPG_ERR_ASS_WRITE_ERROR;
-	}
-    }
 
     pth_cleanup_pop(1);
     pth_cleanup_pop(0);
@@ -1653,7 +1672,7 @@ static gint get_command(assuan_context_t ctx, gchar *line)
 
     if (!req || !*req) {
 	g_strfreev(req);
-	return send_error(ctx, EPWMD_COMMAND_SYNTAX);
+	return send_error(ctx, GPG_ERR_SYNTAX);
     }
 
     n = find_root_element(client->doc, &req, &rc, NULL, 0, FALSE);
@@ -1672,12 +1691,12 @@ static gint get_command(assuan_context_t ctx, gchar *line)
 	return send_error(ctx, rc);
 
     if (!n || !n->children)
-	return send_error(ctx, EPWMD_EMPTY_ELEMENT);
+	return send_error(ctx, GPG_ERR_NO_VALUE);
 
     n = find_text_node(n->children);
 
     if (!n || !n->content || !*n->content)
-	return send_error(ctx, EPWMD_EMPTY_ELEMENT);
+	return send_error(ctx, GPG_ERR_NO_VALUE);
 
     rc = xfer_data(ctx, (gchar *)n->content, xmlStrlen(n->content));
     return send_error(ctx, rc);
@@ -1748,11 +1767,11 @@ static gint realpath_command(assuan_context_t ctx, gchar *line)
 
     if (strchr(line, '\t') != NULL) {
 	if ((req = split_input_line(line, "\t", 0)) == NULL)
-	    return send_error(ctx, EPWMD_COMMAND_SYNTAX);
+	    return send_error(ctx, GPG_ERR_SYNTAX);
     }
     else {
 	if ((req = split_input_line(line, " ", 0)) == NULL)
-	    return send_error(ctx, EPWMD_COMMAND_SYNTAX);
+	    return send_error(ctx, GPG_ERR_SYNTAX);
     }
 
     n = find_root_element(client->doc, &req, &rc, NULL, 0, FALSE);
@@ -1877,7 +1896,7 @@ static gint list_command(assuan_context_t ctx, gchar *line)
 	GString *str;
 
 	if (!total) {
-	    rc = EPWMD_EMPTY_ELEMENT;
+	    rc = GPG_ERR_NO_VALUE;
 	    goto fail;
 	}
 
@@ -1899,7 +1918,7 @@ static gint list_command(assuan_context_t ctx, gchar *line)
 	pth_cleanup_pop(1);
     }
     else
-	rc = EPWMD_EMPTY_ELEMENT;
+	rc = GPG_ERR_NO_VALUE;
 
 fail:
     pth_cleanup_pop(1);
@@ -1921,14 +1940,14 @@ static gint attribute_list(assuan_context_t ctx, gchar **req)
     gpg_error_t rc;
 
     if (!req || !req[0])
-	return EPWMD_COMMAND_SYNTAX;
+	return GPG_ERR_SYNTAX;
 
     if ((path = split_input_line(req[0], "\t", 0)) == NULL) {
 	/*
 	 * The first argument may be only a root element.
 	 */
 	if ((path = split_input_line(req[0], " ", 0)) == NULL)
-	    return EPWMD_COMMAND_SYNTAX;
+	    return GPG_ERR_SYNTAX;
     }
 
     n = find_root_element(client->doc, &path, &rc, NULL, 0, FALSE);
@@ -1976,7 +1995,7 @@ static gint attribute_list(assuan_context_t ctx, gchar **req)
     }
 
     if (!attrlist)
-	return EPWMD_EMPTY_ELEMENT;
+	return GPG_ERR_NO_VALUE;
 
     line = g_strjoinv("\n", attrlist);
 
@@ -2005,14 +2024,14 @@ static gint attribute_delete(struct client_s *client, gchar **req)
     gpg_error_t rc;
 
     if (!req || !req[0] || !req[1])
-	return EPWMD_COMMAND_SYNTAX;
+	return GPG_ERR_SYNTAX;
 
     if ((path = split_input_line(req[1], "\t", 0)) == NULL) {
 	/*
 	 * The first argument may be only a root element.
 	 */
 	if ((path = split_input_line(req[1], " ", 0)) == NULL)
-	    return EPWMD_COMMAND_SYNTAX;
+	    return GPG_ERR_SYNTAX;
     }
 
     /*
@@ -2020,7 +2039,7 @@ static gint attribute_delete(struct client_s *client, gchar **req)
      * root element use DELETE <name>.
      */
     if (!path[1] && xmlStrEqual((xmlChar *)req[0], (xmlChar *)"_name")) {
-	rc = EPWMD_ATTR_SYNTAX;
+	rc = GPG_ERR_SYNTAX;
 	goto fail;
     }
 
@@ -2064,7 +2083,7 @@ again:
     n = find_root_element(client->doc, &src, rc, NULL, 0, FALSE);
 
     if (!n) {
-	if (*rc == EPWMD_ELEMENT_NOT_FOUND) {
+	if (*rc == GPG_ERR_ELEMENT_NOT_FOUND) {
 	    *rc = new_root_element(client->doc, src[0]);
 
 	    if (*rc)
@@ -2129,14 +2148,14 @@ static gpg_error_t target_attribute(struct client_s *client, gchar **req)
     xmlNodePtr n;
 
     if (!req || !req[0] || !req[1])
-	return EPWMD_COMMAND_SYNTAX;
+	return GPG_ERR_SYNTAX;
 
     if ((src = split_input_line(req[0], "\t", 0)) == NULL) {
 	/*
 	 * The first argument may be only a root element.
 	 */
 	if ((src = split_input_line(req[0], " ", 0)) == NULL)
-	    return EPWMD_COMMAND_SYNTAX;
+	    return GPG_ERR_SYNTAX;
     }
 
     if ((dst = split_input_line(req[1], "\t", 0)) == NULL) {
@@ -2144,7 +2163,7 @@ static gpg_error_t target_attribute(struct client_s *client, gchar **req)
 	 * The first argument may be only a root element.
 	 */
 	if ((dst = split_input_line(req[1], " ", 0)) == NULL) {
-	    rc = EPWMD_COMMAND_SYNTAX;
+	    rc = GPG_ERR_SYNTAX;
 	    goto fail;
 	}
     }
@@ -2234,7 +2253,7 @@ static gpg_error_t name_attribute(struct client_s *client, gchar **req)
     n = find_root_element(client->doc, &tmp, &rc, NULL, 0, FALSE);
     g_strfreev(tmp);
 
-    if (rc && rc != EPWMD_ELEMENT_NOT_FOUND)
+    if (rc && rc != GPG_ERR_ELEMENT_NOT_FOUND)
 	return rc;
 
     if (n)
@@ -2244,7 +2263,7 @@ static gpg_error_t name_attribute(struct client_s *client, gchar **req)
      * Whitespace not allowed in root names.
      */
     if (contains_whitespace(req[1]) == TRUE)
-	return EPWMD_ATTR_SYNTAX;
+	return GPG_ERR_SYNTAX;
 
     tmp = g_strdupv(req);
 
@@ -2257,7 +2276,7 @@ static gpg_error_t name_attribute(struct client_s *client, gchar **req)
     g_strfreev(tmp);
 
     if (!n)
-	return EPWMD_ELEMENT_NOT_FOUND;
+	return GPG_ERR_ELEMENT_NOT_FOUND;
 
     return add_attribute(n, "_name", req[1]);
 }
@@ -2275,15 +2294,15 @@ static gint attribute_get(assuan_context_t ctx, gchar **req)
     gpg_error_t rc;
 
     if (!req || !req[0] || !req[1])
-	return EPWMD_COMMAND_SYNTAX;
+	return GPG_ERR_SYNTAX;
 
     if (strchr(req[1], '\t')) {
 	if ((path = split_input_line(req[1], "\t", 0)) == NULL)
-	    return EPWMD_COMMAND_SYNTAX;
+	    return GPG_ERR_SYNTAX;
     }
     else {
 	if ((path = split_input_line(req[1], " ", 0)) == NULL)
-	    return EPWMD_COMMAND_SYNTAX;
+	    return GPG_ERR_SYNTAX;
     }
 
     n = find_root_element(client->doc, &path, &rc, NULL, 0, FALSE);
@@ -2302,7 +2321,7 @@ static gint attribute_get(assuan_context_t ctx, gchar **req)
     g_strfreev(path);
     
     if ((a = xmlGetProp(n, (xmlChar *)req[0])) == NULL)
-	return EPWMD_ATTR_NOT_FOUND;
+	return GPG_ERR_NOT_FOUND;
 
     pth_cleanup_push(xmlFree, a);
 
@@ -2331,7 +2350,7 @@ static gint attribute_set(struct client_s *client, gchar **req)
     xmlNodePtr n;
 
     if (!req || !req[0] || !req[1])
-	return EPWMD_COMMAND_SYNTAX;
+	return GPG_ERR_SYNTAX;
 
     /*
      * Reserved attribute names.
@@ -2352,7 +2371,7 @@ static gint attribute_set(struct client_s *client, gchar **req)
 	 * The first argument may be only a root element.
 	 */
 	if ((path = split_input_line(req[1], " ", 0)) == NULL)
-	    return EPWMD_COMMAND_SYNTAX;
+	    return GPG_ERR_SYNTAX;
     }
 
     n = find_root_element(client->doc, &path, &rc, NULL, 0, FALSE);
@@ -2393,7 +2412,7 @@ static gint attr_command(assuan_context_t ctx, gchar *line)
 
     if (!req || !req[0] || !req[1]) {
 	g_strfreev(req);
-	return send_error(ctx, EPWMD_COMMAND_SYNTAX);
+	return send_error(ctx, GPG_ERR_SYNTAX);
     }
 
     pth_cleanup_push(req_cleanup, req);
@@ -2407,7 +2426,7 @@ static gint attr_command(assuan_context_t ctx, gchar *line)
     else if (g_ascii_strcasecmp(req[0], "LIST") == 0)
 	rc = attribute_list(ctx, req+1);
     else
-	rc = EPWMD_COMMAND_SYNTAX;
+	rc = GPG_ERR_SYNTAX;
 
     pth_cleanup_pop(1);
     return send_error(ctx, rc);
@@ -2421,14 +2440,14 @@ static gint iscached_command(assuan_context_t ctx, gchar *line)
 
     if (!req || !*req) {
 	g_strfreev(req);
-	return send_error(ctx, EPWMD_COMMAND_SYNTAX);
+	return send_error(ctx, GPG_ERR_SYNTAX);
     }
 
     log_write2("ARGS=\"%s\"", line);
 
     if (!valid_filename(req[0])) {
 	g_strfreev(req);
-	return EPWMD_INVALID_FILENAME;
+	return GPG_ERR_INV_VALUE;
     }
 
     gcry_md_hash_buffer(GCRY_MD_MD5, md5file, req[0], strlen(req[0]));
@@ -2475,7 +2494,7 @@ static gint iscached_command(assuan_context_t ctx, gchar *line)
     }
 
     g_free(path);
-    return send_error(ctx, EPWMD_CACHE_NOT_FOUND);
+    return send_error(ctx, GPG_ERR_NOT_FOUND);
 }
 
 static gint clearcache_command(assuan_context_t ctx, gchar *line)
@@ -2499,7 +2518,7 @@ static gint clearcache_command(assuan_context_t ctx, gchar *line)
 
     if (cache_clear(md5file, 1) == FALSE) {
 	CACHE_UNLOCK;
-	return send_error(ctx, EPWMD_CACHE_NOT_FOUND);
+	return send_error(ctx, GPG_ERR_NOT_FOUND);
     }
 
     CACHE_UNLOCK;
@@ -2515,7 +2534,7 @@ static gint cachetimeout_command(assuan_context_t ctx, gchar *line)
 
     if (!req || !*req || !req[1]) {
 	g_strfreev(req);
-	return send_error(ctx, EPWMD_COMMAND_SYNTAX);
+	return send_error(ctx, GPG_ERR_SYNTAX);
     }
 
     errno = 0;
@@ -2523,7 +2542,7 @@ static gint cachetimeout_command(assuan_context_t ctx, gchar *line)
 
     if (errno != 0 || *p != 0 || timeout < -1) {
 	g_strfreev(req);
-	return send_error(ctx, EPWMD_COMMAND_SYNTAX);
+	return send_error(ctx, GPG_ERR_SYNTAX);
     }
 
     gcry_md_hash_buffer(GCRY_MD_MD5, md5file, req[0], strlen(req[0]));
@@ -2531,7 +2550,7 @@ static gint cachetimeout_command(assuan_context_t ctx, gchar *line)
 
     if (cache_set_timeout(md5file, timeout) == FALSE) {
 	CACHE_UNLOCK;
-	return send_error(ctx, EPWMD_CACHE_NOT_FOUND);
+	return send_error(ctx, GPG_ERR_NOT_FOUND);
     }
 
     CACHE_UNLOCK;
@@ -2577,7 +2596,7 @@ static gint getconfig_command(assuan_context_t ctx, gchar *line)
     }
 
     if (fp && !valid_filename(fp))
-	return send_error(ctx, EPWMD_INVALID_FILENAME);
+	return send_error(ctx, GPG_ERR_INV_VALUE);
 
     paramp = g_ascii_strdown(paramp, -1);
 
@@ -2707,7 +2726,7 @@ static gint xpath_command(assuan_context_t ctx, gchar *line)
 	return send_error(ctx, GPG_ERR_NOT_IMPLEMENTED);
 
     if (!line || !*line)
-	return send_error(ctx, EPWMD_COMMAND_SYNTAX);
+	return send_error(ctx, GPG_ERR_SYNTAX);
 
     memset(&xpath, 0, sizeof(struct xpath_s));
 
@@ -2731,7 +2750,7 @@ static gint xpath_command(assuan_context_t ctx, gchar *line)
     }
 
     if (xmlXPathNodeSetIsEmpty(xpath.result->nodesetval)) {
-	rc = EPWMD_ELEMENT_NOT_FOUND;
+	rc = GPG_ERR_ELEMENT_NOT_FOUND;
 	goto fail;
     }
 
@@ -2741,7 +2760,7 @@ static gint xpath_command(assuan_context_t ctx, gchar *line)
     if (rc)
 	goto fail;
     else if (!xpath.req[1] && !xmlBufferLength(xpath.buf)) {
-	rc = EPWMD_EMPTY_ELEMENT;
+	rc = GPG_ERR_NO_VALUE;
 	goto fail;
     }
     else if (xpath.req[1])
@@ -2772,7 +2791,7 @@ static gint xpathattr_command(assuan_context_t ctx, gchar *line)
 	return send_error(ctx, GPG_ERR_NOT_IMPLEMENTED);
 
     if (!line || !*line)
-	return send_error(ctx, EPWMD_COMMAND_SYNTAX);
+	return send_error(ctx, GPG_ERR_SYNTAX);
 
     memset(&xpath, 0, sizeof(struct xpath_s));
 
@@ -2823,7 +2842,7 @@ static gint xpathattr_command(assuan_context_t ctx, gchar *line)
     }
 
     if (xmlXPathNodeSetIsEmpty(xpath.result->nodesetval)) {
-	rc = EPWMD_ELEMENT_NOT_FOUND;
+	rc = GPG_ERR_ELEMENT_NOT_FOUND;
 	goto fail;
     }
 
@@ -2855,13 +2874,13 @@ static gint import_command_finalize(gpointer data, gint assuan_rc, guchar *line,
     xfree(line);
 
     if (!req || !*req)
-	return EPWMD_COMMAND_SYNTAX;
+	return GPG_ERR_SYNTAX;
 
     content = req[0];
     path = split_input_line(req[1], "\t", 0);
 
     if (!content || !*content) {
-	rc = EPWMD_COMMAND_SYNTAX;
+	rc = GPG_ERR_SYNTAX;
 	goto fail;
     }
 
@@ -2905,7 +2924,7 @@ static gint import_command_finalize(gpointer data, gint assuan_rc, guchar *line,
 	xmlFree(a);
 	n = find_root_element(client->doc, &path, &rc, NULL, 0, FALSE);
 
-	if (rc && rc != EPWMD_ELEMENT_NOT_FOUND) {
+	if (rc && rc != GPG_ERR_ELEMENT_NOT_FOUND) {
 	    g_strfreev(path_orig);
 	    goto fail;
 	}
@@ -2913,7 +2932,7 @@ static gint import_command_finalize(gpointer data, gint assuan_rc, guchar *line,
 	if (!rc) {
 	    n = find_elements(client->doc, n->children, path+1, &rc, NULL, NULL, NULL, FALSE, 0, NULL, TRUE);
 
-	    if (rc && rc != EPWMD_ELEMENT_NOT_FOUND) {
+	    if (rc && rc != GPG_ERR_ELEMENT_NOT_FOUND) {
 		g_strfreev(path_orig);
 		goto fail;
 	    }
@@ -2929,15 +2948,15 @@ static gint import_command_finalize(gpointer data, gint assuan_rc, guchar *line,
 	g_strfreev(path);
 	path = path_orig;
 
-	if (rc == EPWMD_ELEMENT_NOT_FOUND) {
+	if (rc == GPG_ERR_ELEMENT_NOT_FOUND) {
 	    n = create_element_path(client, &path, &rc);
 
 	    if (rc)
 		goto fail;
 	}
 
-	copy = xmlCopyNode(root, 1);
-	n = xmlAddChild(n, copy);
+	copy = xmlCopyNodeList(root);
+	n = xmlAddChildList(n, copy);
 
 	if (!n)
 	    rc = EPWMD_LIBXML_ERROR;
@@ -2945,14 +2964,14 @@ static gint import_command_finalize(gpointer data, gint assuan_rc, guchar *line,
     else {
 	/* Check if the content root element can create a DTD root element. */
 	if (!xmlStrEqual((xmlChar *)"element", root->name)) {
-	    rc = EPWMD_COMMAND_SYNTAX;
+	    rc = GPG_ERR_SYNTAX;
 	    goto fail;
 	}
 
 	xmlChar *a;
 
 	if ((a = xmlGetProp(root, (xmlChar *)"_name")) == NULL) {
-	    rc = EPWMD_COMMAND_SYNTAX;
+	    rc = GPG_ERR_SYNTAX;
 	    goto fail;
 	}
 
@@ -2962,7 +2981,7 @@ static gint import_command_finalize(gpointer data, gint assuan_rc, guchar *line,
 
 	if (!valid_xml_element((xmlChar *)tmp) || literal) {
 	    g_free(tmp);
-	    rc = EPWMD_INVALID_ELEMENT;
+	    rc = GPG_ERR_INV_VALUE;
 	    goto fail;
 	}
 
@@ -2975,7 +2994,7 @@ static gint import_command_finalize(gpointer data, gint assuan_rc, guchar *line,
 	g_free(tmp);
 	n = find_root_element(client->doc, &path, &rc, NULL, 0, TRUE);
 
-	if (rc && rc != EPWMD_ELEMENT_NOT_FOUND) {
+	if (rc && rc != GPG_ERR_ELEMENT_NOT_FOUND) {
 	    rc = EPWMD_LIBXML_ERROR;
 	    goto fail;
 	}
@@ -2989,7 +3008,7 @@ static gint import_command_finalize(gpointer data, gint assuan_rc, guchar *line,
 	rc = 0;
 	xmlSetProp(root, (xmlChar *)"_name", (xmlChar *)path[0]);
 	n = xmlCopyNode(root, 1);
-	n = xmlAddChild(xmlDocGetRootElement(client->doc), n);
+	n = xmlAddChildList(xmlDocGetRootElement(client->doc), n);
     }
 
 fail:
@@ -3314,7 +3333,7 @@ static gint rename_command(assuan_context_t ctx, gchar *line)
 
     if (!req || !req[0] || !req[1]) {
 	g_strfreev(req);
-	return send_error(ctx, EPWMD_COMMAND_SYNTAX);
+	return send_error(ctx, GPG_ERR_SYNTAX);
     }
 
     dst = req[1];
@@ -3322,7 +3341,7 @@ static gint rename_command(assuan_context_t ctx, gchar *line)
 
     if (!valid_xml_element((xmlChar *)dst)) {
 	g_strfreev(req);
-	return EPWMD_INVALID_ELEMENT;
+	return GPG_ERR_INV_VALUE;
     }
 
     if (strchr(req[0], '\t'))
@@ -3331,7 +3350,7 @@ static gint rename_command(assuan_context_t ctx, gchar *line)
 	src = split_input_line(req[0], " ", -1);
 
     if (!src || !*src) {
-	rc = EPWMD_COMMAND_SYNTAX;
+	rc = GPG_ERR_SYNTAX;
 	goto fail;
     }
 
@@ -3381,7 +3400,7 @@ static gint rename_command(assuan_context_t ctx, gchar *line)
     strv_printf(&tmp, "!%s", dst);
     ndst = find_root_element(client->doc, &tmp, &rc, NULL, 0, FALSE);
 
-    if (!ndst && rc && rc != EPWMD_ELEMENT_NOT_FOUND) {
+    if (!ndst && rc && rc != GPG_ERR_ELEMENT_NOT_FOUND) {
 	g_strfreev(tmp);
 	goto fail;
     }
@@ -3392,7 +3411,7 @@ static gint rename_command(assuan_context_t ctx, gchar *line)
 
     g_strfreev(tmp);
 
-    if (!ndst && rc && rc != EPWMD_ELEMENT_NOT_FOUND)
+    if (!ndst && rc && rc != GPG_ERR_ELEMENT_NOT_FOUND)
 	goto fail;
 
     rc = 0;
@@ -3437,7 +3456,7 @@ static gint copy_command(assuan_context_t ctx, gchar *line)
 
     if (!req || !req[0] || !req[1]) {
 	g_strfreev(req);
-	return send_error(ctx, EPWMD_COMMAND_SYNTAX);
+	return send_error(ctx, GPG_ERR_SYNTAX);
     }
 
     if (strchr(req[0], '\t'))
@@ -3446,7 +3465,7 @@ static gint copy_command(assuan_context_t ctx, gchar *line)
 	src = split_input_line(req[0], " ", -1);
 
     if (!src || !*src) {
-	rc = EPWMD_COMMAND_SYNTAX;
+	rc = GPG_ERR_SYNTAX;
 	goto fail;
     }
 
@@ -3456,7 +3475,7 @@ static gint copy_command(assuan_context_t ctx, gchar *line)
 	dst = split_input_line(req[1], " ", -1);
 
     if (!dst || !*dst) {
-	rc = EPWMD_COMMAND_SYNTAX;
+	rc = GPG_ERR_SYNTAX;
 	goto fail;
     }
 
@@ -3475,7 +3494,7 @@ static gint copy_command(assuan_context_t ctx, gchar *line)
 	ndst = find_elements(client->doc, ndst->children, dst+1, &rc, NULL,
 		NULL, NULL, FALSE, 0, NULL, FALSE);
 
-    if (!ndst && rc != EPWMD_ELEMENT_NOT_FOUND)
+    if (!ndst && rc != GPG_ERR_ELEMENT_NOT_FOUND)
 	goto fail;
 
     new = xmlCopyNodeList(nsrc);
@@ -3508,12 +3527,31 @@ static gint copy_command(assuan_context_t ctx, gchar *line)
     }
 
     xmlNodePtr n = ndst->children;
-    xmlUnlinkNode(ndst->children);
+    xmlUnlinkNode(n);
     xmlFreeNodeList(n);
-    n = xmlCopyNodeList(new->children);
-    xmlFreeNodeList(new);
     ndst->children = NULL;
-    xmlAddChild(ndst, n);
+
+    if (!new->children) {
+	xmlUnlinkNode(new);
+	xmlFreeNodeList(new);
+	goto fail;
+    }
+
+    n = xmlCopyNodeList(new->children);
+
+    if (!n) {
+	rc = GPG_ERR_ENOMEM;
+	goto fail;
+    }
+
+    xmlUnlinkNode(new);
+    xmlFreeNodeList(new);
+    n = xmlAddChildList(ndst, n);
+
+    if (!n) {
+	rc = GPG_ERR_ENOMEM;
+	goto fail;
+    }
 
 fail:
     if (req)
@@ -3540,7 +3578,7 @@ static gint move_command(assuan_context_t ctx, gchar *line)
 
     if (!req || !req[0] || !req[1]) {
 	g_strfreev(req);
-	return send_error(ctx, EPWMD_COMMAND_SYNTAX);
+	return send_error(ctx, GPG_ERR_SYNTAX);
     }
 
     if (strchr(req[0], '\t'))
@@ -3549,7 +3587,7 @@ static gint move_command(assuan_context_t ctx, gchar *line)
 	src = split_input_line(req[0], " ", -1);
 
     if (!src || !*src) {
-	rc = EPWMD_COMMAND_SYNTAX;
+	rc = GPG_ERR_SYNTAX;
 	goto fail;
     }
 
@@ -3584,7 +3622,7 @@ static gint move_command(assuan_context_t ctx, gchar *line)
 	}
     }
 
-    if (rc && rc != EPWMD_ELEMENT_NOT_FOUND)
+    if (rc && rc != GPG_ERR_ELEMENT_NOT_FOUND)
 	goto fail;
 
     rc = 0;
@@ -3599,10 +3637,9 @@ static gint move_command(assuan_context_t ctx, gchar *line)
 	    if (dup == nsrc)
 		goto fail;
 	    
-	    /* Moving a child element of the source path to the document root
-	     * node that has the same name. */
 	    if (ndst == xmlDocGetRootElement(client->doc)) {
 		xmlNodePtr n = nsrc;
+		gboolean match = FALSE;
 
 		while (n->parent && n->parent != ndst)
 		    n = n->parent;
@@ -3611,18 +3648,22 @@ static gint move_command(assuan_context_t ctx, gchar *line)
 		xmlChar *b = node_has_attribute(nsrc, (xmlChar *)"_name");
 
 		if (xmlStrEqual(a, b)) {
-		    xmlFree(a);
-		    xmlFree(b);
-		    rc = GPG_ERR_CONFLICT;
-		    goto fail;
+		    match = TRUE;
+		    xmlUnlinkNode(nsrc);
+		    xmlUnlinkNode(n);
+		    xmlFreeNodeList(n);
 		}
 
 		xmlFree(a);
 		xmlFree(b);
-	    }
 
-	    xmlUnlinkNode(dup);
-	    xmlFreeNodeList(dup);
+		if (!match) {
+		    xmlUnlinkNode(dup);
+		    xmlFreeNodeList(dup);
+		}
+	    }
+	    else
+		xmlUnlinkNode(dup);
 	}
     }
 
@@ -3633,7 +3674,10 @@ static gint move_command(assuan_context_t ctx, gchar *line)
 	goto fail;
 
     xmlUnlinkNode(nsrc);
-    xmlAddChild(ndst, nsrc);
+    ndst = xmlAddChildList(ndst, nsrc);
+
+    if (!ndst)
+	rc = GPG_ERR_ENOMEM;
 
 fail:
     if (req)
@@ -3971,13 +4015,13 @@ decompress:
 	if (zrc == Z_MEM_ERROR)
 	    return gpg_error_from_errno(ENOMEM);
 	else
-	    return EPWMD_BADKEY; // Not a valid gzip header. Must be a bad key.
+	    return GPG_ERR_INV_PASSPHRASE; // Not a valid gzip header. Must be a bad key.
     }
 
     if (g_strncasecmp(crypto->outbuf, "<?xml ", 6) != 0) {
 	gcry_free(crypto->outbuf);
 	crypto->outbuf = NULL;
-	return EPWMD_BADKEY;
+	return GPG_ERR_INV_PASSPHRASE;
     }
 
     if (ctx) {
