@@ -270,8 +270,8 @@ static gpg_error_t set_pinentry_strings(struct pinentry_s *pin, gint which)
 
     if (!pin->desc && !which)
 	pin->desc = g_strdup_printf(pin->which == PINENTRY_OPEN ?
-		N_("A passphrase is required to open the file \"%s\". Please%%0Aenter the passphrase below.") :
-		N_("A passphrase is required to save to the file \"%s\". Please%%0Aenter the passphrase below."),
+		N_("A passphrase is required to open the file \"%s\". Please enter the passphrase below.") :
+		N_("A passphrase is required to save to the file \"%s\". Please enter the passphrase below."),
 		pin->filename);
 
     if (which == 2)
