@@ -804,7 +804,7 @@ static gpg_error_t read_password(const gchar *filename,
 
 	if (strcmp(crypto->tkey, crypto->tkey2)) {
 	    log_write(N_("%s: Passphrase mismatch"), filename);
-	    return EPWMD_BADKEY;
+	    return GPG_ERR_INV_PASSPHRASE;
 	}
 
 	gcry_md_hash_buffer(GCRY_MD_SHA256, key, crypto->tkey,
